/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 * 
 * Copyright (c) 2007 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the terms of either the GNU Lesser
 * General Public License Version 2.1 only ("LGPL") or the Common Development and
 * Distribution License ("CDDL")(collectively, the "License"). You may not use this
 * file except in compliance with the License. You can obtain a copy of the CDDL at
 * http://www.opensource.org/licenses/cddl1.php and a copy of the LGPLv2.1 at
 * http://www.opensource.org/licenses/lgpl-license.php. See the License for the 
 * specific language governing permissions and limitations under the License. When
 * distributing the software, include this License Header Notice in each file and
 * include the full text of the License in the License file as well as the
 * following notice:
 * 
 * NOTICE PURSUANT TO SECTION 9 OF THE COMMON DEVELOPMENT AND DISTRIBUTION LICENSE
 * (CDDL)
 * For Covered Software in this distribution, this License shall be governed by the
 * laws of the State of California (excluding conflict-of-law provisions).
 * Any litigation relating to this License shall be subject to the jurisdiction of
 * the Federal Courts of the Northern District of California and the state courts
 * of the State of California, with venue lying in Santa Clara County, California.
 * 
 * Contributor(s):
 * 
 * If you wish your version of this file to be governed by only the CDDL or only
 * the LGPL Version 2.1, indicate your decision by adding "[Contributor]" elects to
 * include this software in this distribution under the [CDDL or LGPL Version 2.1]
 * license." If you don't indicate a single choice of license, a recipient has the
 * option to distribute your version of this file under either the CDDL or the LGPL
 * Version 2.1, or to extend the choice of license to its licensees as provided
 * above. However, if you add LGPL Version 2.1 code and therefore, elected the LGPL
 * Version 2 license, then the option applies only if the new code is made subject
 * to such option by the copyright holder. 
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "imi_options.h"
#include "imi_keys.h"

CIMIOptions::~CIMIOptions()
{
}

/*
CIMIOptions* CIMIOptions::clone() const
{
    return new CIMIOptions();
}

bool CIMIOptions::operator==(const CIMIOptions* p2) const
{
    return true;
}

bool CIMIOptions::copy(const CIMIOptions* p2)
{
    return true;
}
*/

CSunpinyinOptions::CSunpinyinOptions() :
        m_ViewType(0), m_MemoryPower(3), m_GBK(1), m_ContextRanking(1),
        m_MinusAsPageUp(1), m_CommaAsPageUp(0), m_BracketAsPageUp(1),
        m_LayoutVeritcal(0), m_Fuzzy(0), m_Fuzzy_zh(1), m_Fuzzy_ch(1),
        m_Fuzzy_sh(1), m_Fuzzy_ln(1), m_Fuzzy_fh(1), m_Fuzzy_ang(1),
        m_Fuzzy_eng(1), m_Fuzzy_ing(1), m_CandiWindowSize(6)
{
}


CSunpinyinOptions::~CSunpinyinOptions()
{
}

CIMIOptions* CSunpinyinOptions::clone() const
{
    #ifdef DEBUG
        printf("Sunpinyin Options clone itself...");
        fflush(stdout);
    #endif
    CSunpinyinOptions* pnew = new CSunpinyinOptions();
    *pnew = *this;
    #ifdef DEBUG
        printf("allocated @0X%X\n", pnew);
        fflush(stdout);
    #endif
    return pnew;
}

bool CSunpinyinOptions::copy(const CIMIOptions* p2)
{
    const CSunpinyinOptions* p = dynamic_cast<const CSunpinyinOptions* >(p2);

    if (p) {
        *this = *p;
        return true;
    }
    return false;
}

bool CSunpinyinOptions::operator==(const CIMIOptions* p2) const
{
    const CSunpinyinOptions* p = dynamic_cast<const CSunpinyinOptions* >(p2);

    if (p) {
        return ((m_ViewType         == p->m_ViewType) &&
                (m_GBK              == p->m_GBK) &&
                (m_MinusAsPageUp    == p->m_MinusAsPageUp) &&
                (m_CommaAsPageUp    == p->m_CommaAsPageUp) &&
                (m_BracketAsPageUp  == p->m_BracketAsPageUp) &&
                (m_CandiWindowSize  == p->m_CandiWindowSize)
#if 0
                &&
                (m_MemoryPower      == p->m_MemoryPower) &&
                (m_ContextRanking   == p->m_ContextRanking) &&
                (m_LayoutVeritcal   == p->m_LayoutVeritcal) &&
                (m_Fuzzy            == p->m_Fuzzy) &&
                (m_Fuzzy_zh         == p->m_Fuzzy_zh) &&
                (m_Fuzzy_ch         == p->m_Fuzzy_ch) &&
                (m_Fuzzy_sh         == p->m_Fuzzy_sh) &&
                (m_Fuzzy_ln         == p->m_Fuzzy_ln) &&
                (m_Fuzzy_fh         == p->m_Fuzzy_fh) &&
                (m_Fuzzy_ang        == p->m_Fuzzy_ang) &&
                (m_Fuzzy_eng        == p->m_Fuzzy_eng) &&
                (m_Fuzzy_ing        == p->m_Fuzzy_ing)
#endif
                );
    }
    return false;
}

bool CSunpinyinOptions::isPageUpKey(unsigned keycode, unsigned keychar, unsigned mask)
{
    if ((mask & (IM_SHIFT_MASK | IM_CTRL_MASK | IM_ALT_MASK)) == 0) {
        return ((keycode == IM_VK_PAGE_UP) ||
                (m_MinusAsPageUp && keychar == IM_VK_MINUS) ||
                (m_CommaAsPageUp && keychar == IM_VK_COMMA) ||
                (m_BracketAsPageUp && keychar == IM_VK_OPEN_BRACKET));
    }
    return false;
}

bool CSunpinyinOptions::isPageDnKey(unsigned keycode, unsigned keychar, unsigned mask)
{
    if ((mask & (IM_SHIFT_MASK | IM_CTRL_MASK | IM_ALT_MASK)) == 0) {
        return ((keycode ==IM_VK_PAGE_DOWN) ||
                (m_MinusAsPageUp && keychar == IM_VK_EQUALS) ||
                (m_CommaAsPageUp && keychar == IM_VK_PERIOD) ||
                (m_BracketAsPageUp && keychar == IM_VK_CLOSE_BRACKET));
    }
    return false;
}
