/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 * 
 * Copyright (c) 2007 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the terms of either the GNU Lesser
 * General Public License Version 2.1 only ("LGPL") or the Common Development and
 * Distribution License ("CDDL")(collectively, the "License"). You may not use this
 * file except in compliance with the License. You can obtain a copy of the CDDL at
 * http://www.opensource.org/licenses/cddl1.php and a copy of the LGPLv2.1 at
 * http://www.opensource.org/licenses/lgpl-license.php. See the License for the 
 * specific language governing permissions and limitations under the License. When
 * distributing the software, include this License Header Notice in each file and
 * include the full text of the License in the License file as well as the
 * following notice:
 * 
 * NOTICE PURSUANT TO SECTION 9 OF THE COMMON DEVELOPMENT AND DISTRIBUTION LICENSE
 * (CDDL)
 * For Covered Software in this distribution, this License shall be governed by the
 * laws of the State of California (excluding conflict-of-law provisions).
 * Any litigation relating to this License shall be subject to the jurisdiction of
 * the Federal Courts of the Northern District of California and the state courts
 * of the State of California, with venue lying in Santa Clara County, California.
 * 
 * Contributor(s):
 * 
 * If you wish your version of this file to be governed by only the CDDL or only
 * the LGPL Version 2.1, indicate your decision by adding "[Contributor]" elects to
 * include this software in this distribution under the [CDDL or LGPL Version 2.1]
 * license." If you don't indicate a single choice of license, a recipient has the
 * option to distribute your version of this file under either the CDDL or the LGPL
 * Version 2.1, or to extend the choice of license to its licensees as provided
 * above. However, if you add LGPL Version 2.1 code and therefore, elected the LGPL
 * Version 2 license, then the option applies only if the new code is made subject
 * to such option by the copyright holder. 
 */

#ifndef SUNPY_IMI_SESSION_VIEW_H
#define SUNPY_IMI_SESSION_VIEW_H

#include "portability.h"

#include "imi_options.h"

#include "imi_context.h"
#include "imi_winHandler.h"
#include "imi_uiobjects.h"

class CIMIView;

class CIMIViewFactory {
    static int          sm_mapInit;
public:
    enum {
        SVT_NONE = 0,
        SVT_MODERN,
        SVT_CLASSIC
    };

    static CIMIView *
    createView(int viewType);
};

class CIMIView {
public:
    enum {
        KEYEVENT_USED =  (1),
        PREEDIT_MASK =   (1<<2),
        CANDIDATE_MASK = (1<<3)
    };

public:
    static int s_CandiWindowSize;

public:
    CIMIView();

    /**{*/
    CIMIOptions* getPreference(void)
        { return m_pPref; }

    void attachWinHandler(CIMIWinHandler* wh)
            { mp_winHandler = wh; }

    CIMIWinHandler* getWinHandler(void)
            { return mp_winHandler; }

    CIMIContext* getIC(void)
            { return m_pIC; }
    /**}*/

    /*@{*/
    virtual ~CIMIView();

    virtual int  getViewType(void);

    virtual void setPreference(const CIMIOptions* pp);

    virtual void attachIC(CIMIContext* pIC);

    virtual unsigned clearIC(void);

    virtual void updateWindows(unsigned int mask);

    virtual void setStatusAttrValue(int key, int value);

    virtual int  onKeyEvent(unsigned keycode, unsigned keyvalue, unsigned modifier);

    virtual int  onCandidatePageRequest(int pgno, bool relative); //pgno == -1, relative == false means last page

    virtual int  onCandidateSelectRequest(int index);

    virtual int  getStatusAttrValue(int key);
    /*@}*/

    /* Functions to retrieve the Preedit/Candidate list */
    /*@{*/
    virtual void getInputString(wstring& s);

    virtual void getPreeditString(IPreeditString& ps);

    virtual void getCandidateList(ICandidateList& cl, int start, int size);
    /*@}*/

protected:
    /*@{*/
    /** Punc include . , ; : ? " ... */
    static bool
    isPuncChar(TWCHAR key);

    /** simbol is those not punc, but has full wide mode */
    static bool
    isSimbolChar(TWCHAR key);

    /** Whether or not the wide char wc is a end-of-senetence puncuation */
    static bool
    isTermPuncWide(TWCHAR wc);
    /*@}*/

    /*@{*/
    /** Map the half Punc char into its full wide char*/
    TWCHAR
    getFullPunc(TWCHAR key);

    /** Map the half simbol char into its full wide char*/
    TWCHAR
    getFullSimbol(TWCHAR key);
    /*@}*/

protected:
    /*@{*/
    CIMIContext        *m_pIC;
    CIMIWinHandler*     mp_winHandler;
    /*@}*/

    /*@{*/
    bool                m_CN;
    bool                m_FullPunc;
    bool                m_FullSimbol;
    bool                m_bRightFullDblQuote;

    CIMIOptions*        m_pPref;
    /*@}*/
};

#endif
