/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 * 
 * Copyright (c) 2007 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the terms of either the GNU Lesser
 * General Public License Version 2.1 only ("LGPL") or the Common Development and
 * Distribution License ("CDDL")(collectively, the "License"). You may not use this
 * file except in compliance with the License. You can obtain a copy of the CDDL at
 * http://www.opensource.org/licenses/cddl1.php and a copy of the LGPLv2.1 at
 * http://www.opensource.org/licenses/lgpl-license.php. See the License for the 
 * specific language governing permissions and limitations under the License. When
 * distributing the software, include this License Header Notice in each file and
 * include the full text of the License in the License file as well as the
 * following notice:
 * 
 * NOTICE PURSUANT TO SECTION 9 OF THE COMMON DEVELOPMENT AND DISTRIBUTION LICENSE
 * (CDDL)
 * For Covered Software in this distribution, this License shall be governed by the
 * laws of the State of California (excluding conflict-of-law provisions).
 * Any litigation relating to this License shall be subject to the jurisdiction of
 * the Federal Courts of the Northern District of California and the state courts
 * of the State of California, with venue lying in Santa Clara County, California.
 * 
 * Contributor(s):
 * 
 * If you wish your version of this file to be governed by only the CDDL or only
 * the LGPL Version 2.1, indicate your decision by adding "[Contributor]" elects to
 * include this software in this distribution under the [CDDL or LGPL Version 2.1]
 * license." If you don't indicate a single choice of license, a recipient has the
 * option to distribute your version of this file under either the CDDL or the LGPL
 * Version 2.1, or to extend the choice of license to its licensees as provided
 * above. However, if you add LGPL Version 2.1 code and therefore, elected the LGPL
 * Version 2 license, then the option applies only if the new code is made subject
 * to such option by the copyright holder. 
 */

#ifndef SUNPY_IMI_CLASSIC_VIEW_H
#define SUNPY_IMI_CLASSIC_VIEW_H

#include "portability.h"

#include "imi_view.h"

class CIMIClassicView : public CIMIView {
public:
    CIMIClassicView();

    virtual ~CIMIClassicView();

    virtual int  getViewType(void);

    virtual void attachIC(CIMIContext* pIC);

    virtual unsigned clearIC(void);

    virtual void updateWindows(unsigned int mask);

    virtual int  onKeyEvent(unsigned keycode, unsigned keyvalue, unsigned modifier);

    virtual int  onCandidatePageRequest(int pgno, bool relative);

    virtual int  onCandidateSelectRequest(int index);

    /*@{*/
    /**
    * get preedit, where from the beginning of the skeleton to the m_candiBone,
    * give the best converted HANZI string. from the m_candiBone (include) to
    * the end of the skeleton, give the original PINYIN string.
    */
    virtual void getPreeditString(IPreeditString& ps);

    virtual void getCandidateList(ICandidateList& cl, int start, int size);
    /*@}*/


protected:
    void
    pressNormalKey(unsigned keyvalue, unsigned int& mask);

    /**
     * Select the idx-th candidate on user's request. move the current
     * Bone forwards to next bone, re-get candidate list. So, all
     * window area (preedit and candidate window) should be update
     */
    void
    makeSelection(int idx, unsigned int& mask);


    /*@{*/
    /**
    * Note only left direction move would cause human selection be discarded
    * @return:  the leftmost bone iterator after operation from which
    *           the new search should begin. If searchAgain = false,
    *           should call explictly to:
    *                m_pIC->searchFrom(leftmost)
    *           value m_pIC->getLastBone() indicat no search are needed or done!
    */
    CSkeletonIter
    moveLeft(unsigned int& mask, bool searchAgain=true);

    CSkeletonIter
    moveLeftSyllable(unsigned int& mask, bool searchAgain=true);

    CSkeletonIter
    moveHome(unsigned int& mask, bool searchAgain=true);

    void
    moveRight(unsigned int& mask);

    void
    moveRightSyllable(unsigned int& mask);

    void
    moveEnd(unsigned int& mask);
    /*@}*/

    /*@{*/
    /**
    * Erase a character near current cursor on the preedit area
    * @param bLeft: true  to erase left char beside cursor.
    *               false to erase right char beside cursor.
    * @param mask: which area should be update after this operation.
    */
    void
    erase(bool bLeft, unsigned int& mask);

    /**
    * Insert a Pinyin character right before the cursor.
    * @param keyvalue: the Pinyin Char to be inserted before cursor.
    * @param mask: which area should be update after this operation.
    */
    void
    insertPinyin(unsigned int keyvalue, unsigned int& mask);

    /**
    * Insert a normal character at the cursor.
    * @param boneType: NODE_ASCII or NODE_PUNC
    * @param keyvalue: the normal char to be inserted, must in [0x20,0x7e].
    * @param mask    : which area should be update after this operation.
    */
    void
    insertNormalChar(int boneType, unsigned orig_value, unsigned keyvalue, unsigned int& mask);

    /**
    * Insert human boundary to current cursor on the preedit area, or change
    * the old boundary at cursor from auto boundary to human boundary.
    * @param mask: which area should be update after this operation.
    */
    void
    insertBoundary(unsigned int& mask);
    /*@}*/

    /*@{*/
    /**
    * commit whole sentence to user.
    * @param bConvert: true  to commit the converted HANZI string.
    *                  false to commit the unconverted PINYIN string.
    */
    void
    doCommit(bool bConvert=true);

    /** commit single character to user, only when skeleton is empty */
    void
    commitChar(TWCHAR ch);

    /**
    * get best sentence from itStart to the end of the skeleton.
    * @param itStart: the starting bone from which convertion starts
    * @param wstr: the HANZI string of the convertion
    * @return : the number of converted words in the wstr string.(ie. do not count
    *           non-pinyin node's result.
    */
    int
    getSentence(wstring& wstr, CSkeletonIter itStart);
    /*@}*/

private:
    /** the bone under the cursor  */
    CSkeletonIter       m_CursorBone;

    /** cursor position inside the bone  */
    int                 m_CursorIdx;

    /** the candidate-list's starting bone.  */
    CSkeletonIter       m_CandiBone;

    /** the candidate list */
    CCandidates         m_CandiList;

    /** the index in whole list of first candidate shown in the candidate bar */
    int                 m_CandiFirst;

    /** tail sentence starting from m_CandiFirst */
    wstring             m_TailSentence;

private:
    void
    getCandidates();

    bool
    cursorAtTail();
};

#endif
