/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 * 
 * Copyright (c) 2007 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the terms of either the GNU Lesser
 * General Public License Version 2.1 only ("LGPL") or the Common Development and
 * Distribution License ("CDDL")(collectively, the "License"). You may not use this
 * file except in compliance with the License. You can obtain a copy of the CDDL at
 * http://www.opensource.org/licenses/cddl1.php and a copy of the LGPLv2.1 at
 * http://www.opensource.org/licenses/lgpl-license.php. See the License for the 
 * specific language governing permissions and limitations under the License. When
 * distributing the software, include this License Header Notice in each file and
 * include the full text of the License in the License file as well as the
 * following notice:
 * 
 * NOTICE PURSUANT TO SECTION 9 OF THE COMMON DEVELOPMENT AND DISTRIBUTION LICENSE
 * (CDDL)
 * For Covered Software in this distribution, this License shall be governed by the
 * laws of the State of California (excluding conflict-of-law provisions).
 * Any litigation relating to this License shall be subject to the jurisdiction of
 * the Federal Courts of the Northern District of California and the state courts
 * of the State of California, with venue lying in Santa Clara County, California.
 * 
 * Contributor(s):
 * 
 * If you wish your version of this file to be governed by only the CDDL or only
 * the LGPL Version 2.1, indicate your decision by adding "[Contributor]" elects to
 * include this software in this distribution under the [CDDL or LGPL Version 2.1]
 * license." If you don't indicate a single choice of license, a recipient has the
 * option to distribute your version of this file under either the CDDL or the LGPL
 * Version 2.1, or to extend the choice of license to its licensees as provided
 * above. However, if you add LGPL Version 2.1 code and therefore, elected the LGPL
 * Version 2 license, then the option applies only if the new code is made subject
 * to such option by the copyright holder. 
 */

#ifndef SUNPY_IMI_MODERN_VIEW_H
#define SUNPY_IMI_MODERN_VIEW_H

#include "portability.h"

#include "imi_view.h"

class CIMIModernView : public CIMIView {
public:
    CIMIModernView();

    virtual ~CIMIModernView();

    virtual int  getViewType(void);

    virtual void attachIC(CIMIContext* pIC);

    virtual unsigned clearIC(void);

    virtual void updateWindows(unsigned int mask);

    virtual void setStatusAttrValue(int key, int value);

    virtual int  onKeyEvent(unsigned keycode, unsigned keyvalue, unsigned modifier);

    virtual int  onCandidatePageRequest(int pgno, bool relative);

    virtual int  onCandidateSelectRequest(int index);

    virtual void getPreeditString(IPreeditString& ps);

    virtual void getCandidateList(ICandidateList& cl, int start, int size);

protected:
    /**
     * cancel current editing syllable, ie., If the syllable is waiting to
     * be inserted to the skeleton as a new bone, clear it. If the syllable
     * is used to modifying a existing bone of the skeleton, unshadow the bone.
     * After this, we enter sentence editing mode. So, both PREEDIT_MASK and
     * CANDIDATE_MASK should be set.
     */
    void
    cancelSyllable(unsigned int& mask);

    /**
     * commit the current editing syllable PINYIN string[0..len). Only
     * when current syllable[0..len) is valid and m_SyllableStr[len..) is
     * non-complete or valid. After commit, m_SyllableStr[len..) become the
     * new value of current syllable string.
     * @param len: m_SyllableStr[0..len) is to be inserted into skeleton
     *             as a valid PINYIN syllable.
     * @param boundaryType: AUOT_BOUNDARY or USER_BOUNDARY here. Specify the
     *              boundary type for the bone to be inserted
     * @param mask: whether or not the windows should be redraw\n
     */
    void
    convertSyllable(int boundaryType, int len, unsigned int& mask);
    /*@}*/


    /**
     * Select the idx-th candidate on user's request. move the current
     * Bone forwards to next bone, re-get candidate list. So, all
     * window area (preedit and candidate window) should be update
     */
    void
    makeSelection(int idx, unsigned int& mask);


    /*@{*/
    /**
     * for simplicity. call m_pIC->modify do most of the work,
     * put the current bone ptr right behind what erased. so do for
     * current candi bone ptr. Put the new m_SyllBone to the position
     * right after boneIt.
     * @param boneIt: the bone to be deleted
     * @param mask whether or not the candi is re-got
     */
    void
    eraseBone(CSkeletonIter boneIt, unsigned int& mask);

    /**
     * insert a Punc or Simbol bone into the skeleton at current position.
     * The core should provide better fast support for this.
     * @param bone: the bone to be added
     * @return whether or not the candidate list is re-got
     */
    void
    insertNonPinyinBone(CBone& bone, unsigned int& mask);
    /*@}*/

    /*@{*/
    /**
    * commit whole sentence to user.
    * @return combination of the area update mask bits
    */
    unsigned
    doCommit(bool bConvert=true);

    /** commit sentence at the left side of the m_SyllBone to user */
    void
    tryCommitLeftSentence(unsigned int& mask);

    /** commit single character to user, only when skeleton is empty */
    void
    commitChar(TWCHAR ch);

    /**
     * get best sentence
     * @param pSyllPos to return the starting position for current syllable
     * @param pSyllSize to return the size of current syllable
     * @return the caret position of the char
     */
    int
    getSentence(wstring& wstr, int* pSyllPos=NULL, int* pSyllSize=NULL);
    /*@}*/

    /*@{*/
    /**
     * Process key when we are in syllable editing mode, ie,
     * the m_SyllableStr is not zero.
     */
    void
    pressKeyOnSyllable(unsigned keycode, unsigned keyvalue, unsigned modifier, unsigned int& mask);

    /**
     * Process key when we are in sentence editing mode, ie,
     * the m_SyllableStr is zero length.
     */
    void
    pressKeyOnSentence(unsigned keycode, unsigned keyvalue, unsigned modifier, unsigned int& mask);

    /**
     * Process non-pinyin character input when we are in sentence
     * editing mode, ie, the m_SyllableStr is zero length. for those
     * non-printable character, return it back to IM System as we do
     * not accept the event (make no change to mask).
     * First, we map Punc and Simbol, and when IC is empty, commit the
     * mapped char. When IC has some content, make a bone and insert
     * it into the skeleton.
     */
    void
    pressNormalKey(unsigned keycode, unsigned keyvalue, unsigned modifier, unsigned int& mask);
    /*@}*/

protected:
    /*@{*/
    /**
     * The syllable value we are now editing upon m_curBone.
     * Note, that when it is not zero lengthed, whether or not
     * EN state can be switched into, or whether or not non-pinyin
     * char could be input should be considered carefully.  By the
     * way, the caret is always positioned at the tail of this string.
     */
    wstring             m_SyllableStr;

    /** Whether the Syllable is from a existing bone, or new to be inserted */
    bool                m_bNewSyll;

    /**
     * Current Syllable which the caret is just before or upon. this is
     * also the candidate-list's starting bone.
     */
    CSkeletonIter       m_SyllBone;

    /** the candidate list */
    CCandidates         m_CandiList;

    /** the index in whole list of first candidate shown in the candidate bar */
    int                 m_CandiFirst;

    /** cursor position inside the current syllable we are now editing */
    int                 m_CursorIdx;

    /*@}*/

};

#endif
